local nvim_lsp = require("lspconfig")

local lsp_attach = function(args)
  return function(client, bufnr)
    if args == nil or args.format == nil or args.format then
      local augroup = vim.api.nvim_create_augroup("LspFormatting", {})
      if client.supports_method("textDocument/formatting") then
        vim.api.nvim_clear_autocmds({ group = augroup, buffer = bufnr })
        vim.api.nvim_create_autocmd("BufWritePre", {
          group = augroup,
          buffer = bufnr,
          callback = function()
            vim.lsp.buf.format()
          end,
        })
      end
    end

    vim.api.nvim_buf_set_option(bufnr, "omnifunc", "v:lua.MiniCompletion.completefunc_lsp")

    local opts = { buffer = bufnr }

    vim.keymap.set("n", "[d", vim.diagnostic.goto_prev)
    vim.keymap.set("n", "]d", vim.diagnostic.goto_next)
    vim.keymap.set("n", "<leader>wd", vim.diagnostic.open_float)
    vim.keymap.set("n", "<leader>ww", vim.diagnostic.setqflist)

    vim.keymap.set("i", "<C-h>", vim.lsp.buf.signature_help, opts)
    vim.keymap.set("n", "K", vim.lsp.buf.hover, opts)
    vim.keymap.set("n", "gd", vim.lsp.buf.definition, opts)
    vim.keymap.set("n", "gD", vim.lsp.buf.declaration, opts)
    vim.keymap.set("n", "gi", vim.lsp.buf.implementation, opts)
    vim.keymap.set("n", "gI", vim.lsp.buf.type_definition, opts)
    vim.keymap.set("n", "gr", vim.lsp.buf.references, opts)
    vim.keymap.set("n", "z=", vim.lsp.buf.code_action, opts)
    vim.keymap.set("n", "zr", vim.lsp.buf.rename, opts)
  end
end

local capabilities = vim.lsp.protocol.make_client_capabilities()

local servers = {
  -- "clangd",
  "clojure_lsp",
  "hls",
  "julials",
  "purescriptls",
  -- "rust_analyzer", -- see config/rust-tools.lua
  "tsserver",
}

for _, lsp in ipairs(servers) do
  nvim_lsp[lsp].setup({
    on_attach = lsp_attach({}),
    capabilities = capabilities,
    flags = { debounce_text_changes = 150 },
  })
end

nvim_lsp.pyright.setup({
  settings = {
    python = {
      analysis = {
        autoImportCompletions = true,
        autoSearchPaths = true,
        diagnosticMode = "openFilesOnly", -- or "workspace"
        typeCheckingMode = "on",
        useLibraryCodeForTypes = true,
      },
      organizeimports = {
        provider = "isort",
      },
    },
  },
  on_attach = lsp_attach({ format = false }),
})

nvim_lsp.racket_langserver.setup({
  filetypes = { "racket" },
  on_attach = lsp_attach({}),
  flags = { debounce_text_changes = 150 },
})

local configs = require("lspconfig.configs")
configs.lsp_wl = {
  default_config = {
    cmd = {
      "wolfram",
      "kernel",
      "-noinit",
      "-noprompt",
      "-nopaclet",
      "-noicon",
      "-nostartuppaclets",
      "-run",
      'Needs["LSPServer`"];LSPServer`StartServer[]',
    },
    filetypes = { "mma" },
    root_dir = nvim_lsp.util.path.dirname,
    autostart = false,
  },
}
nvim_lsp.lsp_wl.setup({
  -- FIXME: better to disable autoformatting from lsp_attach()
  on_attach = function(client, bufnr)
    vim.api.nvim_buf_set_option(bufnr, "omnifunc", "v:lua.MiniCompletion.completefunc_lsp")
    vim.keymap.set("n", "[d", vim.diagnostic.goto_prev)
    vim.keymap.set("n", "]d", vim.diagnostic.goto_next)
    vim.keymap.set("n", "<leader>wd", vim.diagnostic.open_float)
    vim.keymap.set("n", "<leader>ww", vim.diagnostic.setqflist)
    vim.keymap.set("i", "<C-h>", vim.lsp.buf.signature_help, opts)
    vim.keymap.set("n", "K", vim.lsp.buf.hover, opts)
    vim.keymap.set("n", "gd", vim.lsp.buf.definition, opts)
    vim.keymap.set("n", "gD", vim.lsp.buf.declaration, opts)
    vim.keymap.set("n", "gi", vim.lsp.buf.implementation, opts)
    vim.keymap.set("n", "gI", vim.lsp.buf.type_definition, opts)
    vim.keymap.set("n", "gr", vim.lsp.buf.references, opts)
    vim.keymap.set("n", "z=", vim.lsp.buf.code_action, opts)
    vim.keymap.set("n", "zr", vim.lsp.buf.rename, opts)
  end,
})
